package es.uca.xmleye;

/*
 Copyright (C) 2008 Antonio García Domínguez

 This program is free software; you can redistribute it 
 and/or modify it under the terms of the GNU General Public 
 License as published by the Free Software Foundation; 
 either version 2 of the License, or (at your option) any 
 later version.

 This program is distributed in the hope that it will be 
 useful, but WITHOUT ANY WARRANTY; without even the implied 
 warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 PURPOSE. See the GNU General Public License for more 
 details.

 You should have received a copy of the GNU General Public 
 License along with this program; if not, write to the Free 
 Software Foundation, Inc., 59 Temple Place, Suite 330, 
 Boston, MA 02111-1307 USA

 */

import java.util.Locale;
import java.util.ResourceBundle;
import java.io.File;
import java.io.IOException;
import javax.xml.parsers.*;
import javax.xml.xpath.*;
import org.w3c.dom.*;
import org.xml.sax.SAXException;

import es.uca.xmleye.eventos.AlcanceBusqueda;
import es.uca.xmleye.eventos.EventoBusqueda;
import es.uca.xmleye.xpath.*;
import es.uca.xmleye.xslt.ExcepcionTransformacion;
import es.uca.xmleye.xslt.HojaPreprocesado;

import static es.uca.xmleye.eventos.AlcanceBusqueda.*;

/**
 * Modelo de documento de la aplicación.
 */
public class ModeloDocumento {
	// Usado para internacionalizar los mensajes de error
	protected static final ResourceBundle errRes;

	static {
		errRes = ResourceBundle.getBundle("es.uca.xmleye.l10n.ErrorMessages",
				Locale.getDefault());
	}

	// Documento XML
	private Document arbolDOM;
	// Hoja de preprocesado
	private HojaPreprocesado hojaPreproc;
	// Fichero fuente XML y original (posiblemente el mismo)
	private File fichXML, fichOriginal;
	// Directorio base desde el que localizar el fichero fuente
	private String directorioBase;

	/* ------------- CONSTRUCTORES / METODOS PUBLICOS ------------------- */
	public ModeloDocumento(File pFichXML, HojaPreprocesado pHojaPreproc,
			File pOriginal,
			String directorioBase) throws ExcepcionAnalisis,
			ExcepcionTransformacion {
		this.fichXML        = pFichXML;
		this.fichOriginal   = pOriginal;
		this.directorioBase = directorioBase;
		setHojaPreprocesado(pHojaPreproc);
	}

	/**
	 * Obtiene la hoja de preprocesado utilizada.
	 * 
	 * @return Hoja de preprocesado en uso.
	 */
	public HojaPreprocesado getHojaPreprocesado() {
		return hojaPreproc;
	}

	/**
	 * Cambia la hoja de preprocesado, regenerando el documento.
	 * 
	 * @param pHojaPreproc
	 *            Nueva hoja de preprocesado.
	 * @throws ExcepcionTransformacion
	 *             Fallo durante la transformación XSLT.
	 * @throws ExcepcionAnalisis
	 *             Fallo durante el analisis del arbol XML resultante.
	 */
	public void setHojaPreprocesado(HojaPreprocesado pHojaPreproc)
			throws ExcepcionTransformacion, ExcepcionAnalisis {
		try {
			final DocumentBuilder docBuilder = DocumentBuilderFactory
					.newInstance().newDocumentBuilder();
			final Document xmlOriginal = docBuilder.parse(getFicheroXML());

			this.arbolDOM    = pHojaPreproc.preprocesar(xmlOriginal);
			this.hojaPreproc = pHojaPreproc;

			// TODO: convertir esto en una opción configurable, útil para el
			// desarrollador de extensiones. También sería útil para depurar las vistas.
			/*
			try {
				VolcadorXML.volcarXML(this.arbolDOM.getDocumentElement(),
						System.out);
			} catch (Exception ex) {
			}*/

		} catch (ParserConfigurationException pEx) {
			throw new ExcepcionAnalisis(errRes.getString("err_conf_xmlparser"),
					pEx);
		} catch (SAXException saxEx) {
			throw new ExcepcionAnalisis(errRes.getString("err_sax_xmlparser")
					+ "\n" + saxEx.getLocalizedMessage(), saxEx);
		} catch (IOException ioEx) {
			throw new ExcepcionAnalisis(errRes.getString("err_io_xmlparser")
					+ "\n" + ioEx.getLocalizedMessage(), ioEx);
		}
	}

	/**
	 * Devuelve el árbol DOM del documento, ya preprocesado.
	 */
	public Document getArbolDOM() {
		return arbolDOM;
	}
	
	/**
	 * Devuelve el fichero XML del documento.
	 */
	public File getFicheroXML() {
		return fichXML;
	}

	/**
	 * Devuelve el fichero fuente del documento.
	 */
	public File getFicheroOriginal() {
		return fichOriginal;
	}

	/**
	 * Devuelve la hoja de preprocesado usada en el documento.
	 */
	public HojaPreprocesado getHojaPreproc() {
		return hojaPreproc;
	}

	/**
	 * Devuelve el elemento raíz del documento.
	 */
	public Element getRaiz() {
		return (Element) getArbolDOM().getDocumentElement();
	}

	/**
	 * Devuelve la ruta base a usar con rutas relativas.
	 */
	public String getRutaBase() {
		return directorioBase;
	}

	/**
	 * <p>
	 * Realiza una búsqueda a lo largo del árbol, habiendo seleccionado el nodo
	 * <code>nodoSel</code>, mediante los parámetros del evento de petición
	 * de búsqueda e.
	 * </p>
	 * 
	 * @param ev
	 *            Parámetros de la búsqueda.
	 * @param nodoSel
	 *            Nodo seleccionado.
	 * @return Lista de coincidencias.
	 */
	public NodeList buscar(EventoBusqueda ev, Node nodoSel)
			throws ExcepcionBusqueda {
		// Raíz del árbol
		final Element raiz = getRaiz();

		/* Corregimos el alcance de la búsqueda */
		if (ev.getAlcance() == null) {
			ev.setAlcance(ALCANCE_DOCUMENTO);
		}
		if (ev.getAlcance() == ALCANCE_DOCUMENTO) {
			nodoSel = raiz;
			ev.setAlcance(ALCANCE_SIGUIENTES_SUBARBOL);
		}

		/* Tratamos casos en que el alcance sea compuesto */
		switch (ev.getAlcance()) {
		case ALCANCE_SIGUIENTES_SUBARBOL: {
			final EventoBusqueda evSiguientes = ev.clone();
			evSiguientes.setAlcance(ALCANCE_SIGUIENTES);
			final NodeList nodosSiguientes = buscarXPath(evSiguientes, raiz,
					nodoSel);

			final EventoBusqueda evSubtree = ev.clone();
			evSubtree.setAlcance(ALCANCE_SUBARBOL);
			final NodeList nodosSubarbol = buscarXPath(evSubtree, raiz, nodoSel);

			final VectorNodeList listaFinal = new VectorNodeList();
			listaFinal.add(nodosSubarbol);
			listaFinal.add(nodosSiguientes);
			return listaFinal;
		}
		case ALCANCE_ANTERIORES_ANCESTROS: {
			final EventoBusqueda evAnteriores = ev.clone();
			evAnteriores.setAlcance(ALCANCE_ANTERIORES);
			final NodeList nodosAnteriores = buscarXPath(evAnteriores, raiz,
					nodoSel);

			final EventoBusqueda evAncestros = ev.clone();
			evAncestros.setAlcance(ALCANCE_ANCESTROS);
			final NodeList nodosAncestros = buscarXPath(evAncestros, raiz,
					nodoSel);

			final VectorNodeList listaFinal = new VectorNodeList();
			listaFinal.add(nodosAnteriores);
			listaFinal.add(nodosAncestros);
			listaFinal.invertir();
			return listaFinal;
		}
		default:
			return buscarXPath(ev, raiz, nodoSel);
		}
	}

	/* ---------------------- METODOS PRIVADOS ----------------------------- */

	/**
	 * Método de implementación de la búsqueda. No realiza funciones de control,
	 * sólo la búsqueda en sí.
	 * 
	 * @param ev
	 *            Evento delegado por el controlador.
	 * @param findRoot
	 *            Nodo contexto de la búsqueda.
	 * @throw XPathExpressionException Error en la expresión XPath.
	 */
	private NodeList buscarXPath(EventoBusqueda ev, Node raiz, Node nodoSel)
			throws ExcepcionBusqueda {
		// Construimos la clave de búsqueda XPath
		final String exprXPath = crearExprBusquedaXPath(ev, raiz, nodoSel);
		// Creamos el objeto XPath y preparamos las extensiones
		XPath xpath = XPathFactory.newInstance().newXPath();
		xpath.setXPathFunctionResolver(new ResExtensionesXPath());

		try {
			return (NodeList) xpath.evaluate(exprXPath, nodoSel,
					XPathConstants.NODESET);
		} catch (XPathExpressionException xpathEx) {
			throw new ExcepcionBusqueda(errRes
					.getString("err_eval_xpathexpr_1")
					+ exprXPath
					+ errRes.getString("err_eval_xpathexpr_2")
					+ "\n" + xpathEx.getCause(), xpathEx);
		}
	}

	/**
	 * <p>
	 * Construye la expresión XPath a usar en una búsqueda.
	 * </p>
	 * 
	 * <p>
	 * <b>NOTA</b>: este método no considera los alcances compuestos, tales
	 * como <code>ALCANCE_SIGUIENTES_SUBARBOL</code>. Se ha de descomponer
	 * dichos casos en sus componentes simples, aquí
	 * <code>ALCANCE_SIGUIENTES</code> y <code>ALCANCE_SUBARBOL</code>.
	 * </p>
	 * 
	 * @param ev
	 *            Parámetros de la búsqueda.
	 * @param raiz
	 *            Nodo raíz del árbol.
	 * @param nodoSel
	 *            Nodo seleccionado.
	 * @return Expresión XPath a usar posteriormente en buscarXPath().
	 */
	private String crearExprBusquedaXPath(EventoBusqueda ev, Node raiz,
			Node nodoSel) {
		if (ev.getRutaXPath() != null) {
			return ev.getRutaXPath();
		}

		final AlcanceBusqueda alcance = ev.getAlcance();
		final String searchString = ev.getClaveBusqueda();
		final boolean bPalabraCompleta = ev.getPalabraCompleta();
		final boolean bCoincidenMayusMinus = ev.getCoincidenMayusMinus();

		// Ruta a usar en la búsqueda XPath
		String ejeNodo;
		switch (alcance) {
		case ALCANCE_DOCUMENTO:
			ejeNodo = "//*";
			break;
		case ALCANCE_SUBARBOL:
			ejeNodo = "descendant::*";
			break;
		case ALCANCE_SIGUIENTES:
			ejeNodo = "following::*";
			break;
		case ALCANCE_ANTERIORES:
			ejeNodo = "preceding::*";
			break;
		case ALCANCE_ANCESTROS:
		default:
			ejeNodo = "ancestor::*";
			break;
		}

		// Llamada a realizar para obtener el texto del nodo
		String obtencionTexto;
		if (!bCoincidenMayusMinus) {
			obtencionTexto = FuncionMinusculas.getFQN() + "(string(.))";
		} else {
			obtencionTexto = "string(.)";
		}
		// Llamada a realizar para usar la clave de búsqueda
		String cadenaClave;
		if (!bCoincidenMayusMinus) {
			cadenaClave = FuncionMinusculas.getFQN() + "('" + searchString
					+ "')";
		} else {
			cadenaClave = "'" + searchString + "'";
		}

		// Predicado a usar para identificar un nodo
		String predicado;
		if (bPalabraCompleta) {
			predicado = FuncionPalabraCompleta.getFQN() + "(" + obtencionTexto
					+ "," + cadenaClave + ")";
		} else {
			predicado = "contains(" + obtencionTexto + "," + cadenaClave + ")";
		}
		predicado = "text()[" + predicado + "] or @*[" + predicado + "]";

		/*
		 * Para adaptar la búsqueda XPath, que opera sobre el árbol DOM original
		 * sin tener en cuenta los tags de control de visualización que hemos
		 * añadido, a nuestro árbol controlado por datos, debemos tener en
		 * cuenta que: - no se deben obtener nodos que no sean visibles. Dichos
		 * nodos son los que tienen un padre marcado como "hoja", o están
		 * marcados como "ocultos" - un nodo marcado como "hoja" contiene toda
		 * la información de sus descendientes (esperamos que se use en la
		 * visualización, para no confundir al usuario), estén ocultos o no. -
		 * un nodo contiene la información de sus hijos ocultos, y de los
		 * descendientes de éstos.
		 */
		String predFinal = ejeNodo + "[" + predicado
				+ " or (@leaf = '1' and descendant::*[" + predicado + "])"
				+ " or (*[@hidden = '1' and descendant-or-self::*[" + predicado
				+ "]])]" + "[not(@hidden = '1')]"
				+ "[not(ancestor::*[@leaf = '1'])]";

		return predFinal;
	}

}
