package es.uca.xmleye.presentacion;

/*
 Copyright (C) 2008 Antonio García Domínguez

 This program is free software; you can redistribute it 
 and/or modify it under the terms of the GNU General Public 
 License as published by the Free Software Foundation; 
 either version 2 of the License, or (at your option) any 
 later version.

 This program is distributed in the hope that it will be 
 useful, but WITHOUT ANY WARRANTY; without even the implied 
 warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 PURPOSE. See the GNU General Public License for more 
 details.

 You should have received a copy of the GNU General Public 
 License along with this program; if not, write to the Free 
 Software Foundation, Inc., 59 Temple Place, Suite 330, 
 Boston, MA 02111-1307 USA
 */

import java.io.File;
import java.util.Locale;
import java.util.ResourceBundle;
import javax.swing.tree.*;
import org.w3c.dom.*;

import es.uca.xmleye.Editor;
import es.uca.xmleye.ExcepcionAnalisis;
import es.uca.xmleye.ExcepcionBusqueda;
import es.uca.xmleye.ExcepcionEdicion;
import es.uca.xmleye.ModeloDocumento;
import es.uca.xmleye.eventos.EventoBusqueda;
import es.uca.xmleye.gui.arboles.ModArbolDirigidoDatos;
import es.uca.xmleye.util.Contenedor;
import es.uca.xmleye.util.MarcadorClaveXHTML;
import es.uca.xmleye.util.NotificadorCambiosFichero;
import es.uca.xmleye.xslt.ExcepcionTransformacion;
import es.uca.xmleye.xslt.HojaPreprocesado;
import es.uca.xmleye.xslt.HojaVisualizacion;

/**
 * Modelo de presentación de un documento.
 */
public class ModPresentacionDocumento {
	// Expresión regular para marcado de apariciones de la clave de búsqueda
	private final static String DESTACA_REGEXP = "<font class=\"searchkey\">$1</font>";
	// Recursos para localización
	private static final ResourceBundle miscRes;

	static {
		miscRes = ResourceBundle.getBundle("es.uca.xmleye.l10n.MiscMessages",
				Locale.getDefault());
	}

	private ModPresentacionVentanaPrincipal ventanaPrincipal; // Modelo de
	// ventana
	// principal
	private HojaVisualizacion hojaVista; // Hoja de visualización
	private ModeloDocumento modDocumento; // Modelo del documento

	private boolean bNavegable; // ¿Documento navegable?
	private ModArbolDirigidoDatos modArbol; // Modelo del árbol para JTree
	private TreePath seleccion; // Ruta seleccionada
	private String fuenteXHTML; // Código XHTML a visualizar

	private EventoBusqueda ultimaBusqueda; // Última búsqueda realizada
	private NodeList resUltimaBusqueda; // Caché de resultados
	private int indiceUltimoResultado; // Índice del último resultado

	private NotificadorCambiosFichero notCambios; // Notificador de cambios de

	// la fuente

	/* --- CONSTRUCTORES --- */

	/**
	 * Constructor para el caso en que no hay documento abierto.
	 */
	public ModPresentacionDocumento(
			ModPresentacionVentanaPrincipal ventanaPrincipal)
			throws ExcepcionAnalisis, ExcepcionTransformacion {
		this.ventanaPrincipal = ventanaPrincipal;
		this.hojaVista = null;
		this.modDocumento = null;

		this.bNavegable = false;
		this.modArbol = new ModArbolDirigidoDatos();
		this.seleccion = null;

		this.ultimaBusqueda = null;
		this.resUltimaBusqueda = null;

		this.fuenteXHTML = generarFuenteXHTML(this.seleccion);
		this.notCambios = null;
	}

	public ModPresentacionDocumento(ModeloDocumento pMod,
			HojaVisualizacion pHojaVista,
			ModPresentacionVentanaPrincipal ventanaPrincipal)
			throws ExcepcionTransformacion {

		this.ventanaPrincipal = ventanaPrincipal;
		this.hojaVista = pHojaVista;

		setModeloDocumento(pMod);

		this.bNavegable = true;
		this.ultimaBusqueda = null;
		this.resUltimaBusqueda = null;

		// Si se quiere recibir notificaciones, habrá de activarse
		// explícitamente
		this.notCambios = new NotificadorCambiosFichero(pMod
				.getFicheroOriginal(),
				NotificadorCambiosFichero.INTERVALO_OMISION,
				NotificadorCambiosFichero.INACTIVO);
	}

	/* --- METODOS PUBLICOS --- */

	// Accesores para sincronización
	public String getNombreHojaVista() {
		return (hojaVista != null) ? hojaVista.getNombre()
				: ConstantesPreferencias.INICIAL_VISTA_OMISION;
	}

	public String getNombreHojaPreproc() {
		return (modDocumento != null) ? modDocumento.getHojaPreproc()
				.getNombre()
				: ConstantesPreferencias.INICIAL_PREPROCESADO_OMISION;
	}

	public String getNombreDocumento() {
		return (modDocumento != null) ? getFicheroOrigen().getName() : null;
	}

	public TreeModel getModArbol() {
		return modArbol;
	}

	public TreePath getSeleccion() {
		return seleccion;
	}

	public String getFuenteXHTML() {
		return fuenteXHTML;
	}

	public boolean getNavegable() {
		return bNavegable;
	}

	public NotificadorCambiosFichero getNotCambios() {
		return notCambios;
	}

	public String getRutaBase() {
		return (modDocumento != null) ? modDocumento.getRutaBase() : null;
	}

	// Notificadores para eventos
	public synchronized void documentoMostrado() {

	}

	public synchronized void documentoOcultado() {

	}

	public synchronized NodeList solicitudBusqueda(EventoBusqueda e)
			throws ExcepcionBusqueda {
		final Contenedor<String> elem = ventanaPrincipal.inicioTarea(miscRes
				.getString("status_searching"));

		try {
			final Node nodoSel;
			if (getSeleccion().getLastPathComponent() != null) {
				nodoSel = (Node) getSeleccion().getLastPathComponent();
			} else {
				nodoSel = (Node) modArbol.getRoot();
			}

			NodeList res;
			int posSel = buscarPosicion(nodoSel, resUltimaBusqueda);
			if (e.equals(ultimaBusqueda) && posSel != -1) {
				this.indiceUltimoResultado = posSel + 1;
				res = this.resUltimaBusqueda;
			} else {
				this.ultimaBusqueda = e;
				this.resUltimaBusqueda = null;
				this.indiceUltimoResultado = 0;

				res = modDocumento.buscar(e, nodoSel);
			}

			return res;
		} catch (ExcepcionBusqueda ex) {
			throw ex;
		} finally {
			ventanaPrincipal.finTarea(elem);
		}
	}

	public synchronized NodeList repetirBusqueda() {
		return this.resUltimaBusqueda;
	}

	public synchronized void finBusqueda(NodeList resultados, EventoBusqueda e) {
		// Sólo aceptamos los resultados de la búsqueda más reciente
		// (diferenciamos entre referencias distintas - posiblemente
		// tengan orígenes distintos)
		if (e != null && !e.equals(ultimaBusqueda))
			return;

		// Almacenamos los resultados por si se repite la búsqueda
		resUltimaBusqueda = resultados;

		// Cambiamos la selección al primer/siguiente resultado (si lo hay)
		if (resultados != null && resultados.getLength() > 0) {
			if (indiceUltimoResultado >= resultados.getLength()) {
				// Volvemos al primer resultado si no quedan más
				indiceUltimoResultado = 0;
			}
			this.seleccion = modArbol.caminoNodo(resultados
					.item(indiceUltimoResultado++));
		}
	}

	public synchronized String solicitudCambioSeleccion(TreePath nuevaSel)
			throws ExcepcionTransformacion {
		final Contenedor<String> elem = ventanaPrincipal.inicioTarea(miscRes
				.getString("status_transforming"));

		try {
			String res = generarFuenteXHTML(nuevaSel);
			return res;
		} catch (ExcepcionTransformacion ex) {
			throw ex;
		} finally {
			ventanaPrincipal.finTarea(elem);
		}

	}

	public synchronized void finCambioSeleccion(String nuevoXHTML,
			TreePath nuevaSel) {
		this.seleccion = nuevaSel;
		this.fuenteXHTML = nuevoXHTML;
	}

	public synchronized void editarDocumento() throws ExcepcionEdicion {
		File f = getFicheroOrigen();
		String orden = ventanaPrincipal.getOrdenEditor(f);

		if (f != null) {
			Editor.lanzarEditor(f, orden);
		}
	}

	/* --- METODOS DE PAQUETE: SÓLO USAR EN ModPresentacionVentanaPrincipal --- */

	// / Obtención del fichero XML
	synchronized File getFicheroFuente() {
		return this.modDocumento.getFicheroXML();
	}

	// / Obtención del fichero origen
	synchronized File getFicheroOrigen() {
		return this.modDocumento.getFicheroOriginal();
	}

	/**
	 * Notificación por ModPresentacionVentanaPrincipal de cambio de hoja de
	 * visualización.
	 */
	synchronized void cambioHojaVisualizacion(HojaVisualizacion nuevaHoja)
			throws ExcepcionTransformacion {
		this.hojaVista = nuevaHoja;
		this.fuenteXHTML = generarFuenteXHTML(this.seleccion);
	}

	/**
	 * Notificación por ModPresentacionVentanaPrincipal de cambio de hoja de
	 * preprocesado.
	 */
	synchronized void cambioHojaPreprocesado(HojaPreprocesado nuevaHoja)
			throws ExcepcionAnalisis, ExcepcionTransformacion {
		modDocumento.setHojaPreprocesado(nuevaHoja);
		setModeloDocumento(this.modDocumento);
	}

	// / Notificación por ModPresentacionDocumento de cambio de importacion
	// automatica
	synchronized void cambioAutoImportacion(boolean activo) {
		if (activo) {
			notCambios.reanudar();
		} else {
			notCambios.pausar();
		}
	}

	// / Obtención de modelo de documento
	synchronized ModeloDocumento getModeloDocumento() {
		return this.modDocumento;
	}

	// / Cambio de modelo de documento
	synchronized void setModeloDocumento(ModeloDocumento pMod)
			throws ExcepcionTransformacion {
		this.modDocumento = pMod;
		this.modArbol = new ModArbolDirigidoDatos(modDocumento.getArbolDOM());
		this.seleccion = new TreePath(modArbol.getRoot());
		this.fuenteXHTML = generarFuenteXHTML(this.seleccion);
	}

	/* --- METODOS PRIVADOS --- */
	private String generarFuenteXHTML(TreePath sel)
			throws ExcepcionTransformacion {
		if (sel != null && (Element) sel.getLastPathComponent() != null) {
			final Element nodoSel = (Element) sel.getLastPathComponent();
			String xhtml = hojaVista.generarXHTML(nodoSel, modDocumento
					.getArbolDOM());

			if (resUltimaBusqueda != null
					&& buscarPosicion(nodoSel, resUltimaBusqueda) != -1) {
				xhtml = new MarcadorClaveXHTML().marcarApariciones(xhtml,
						ultimaBusqueda.getClaveBusqueda(), DESTACA_REGEXP,
						ultimaBusqueda.getPalabraCompleta(), ultimaBusqueda
								.getCoincidenMayusMinus());
			}

			return xhtml;
		} else {
			return "<html><head></head><body></body></html>";
		}

		// TODO: reintegrar marcado
	}

	/**
	 * Devuelve la posición de aguja en pajar o -1 si no está.
	 */
	private int buscarPosicion(Node aguja, NodeList pajar) {
		if (pajar == null || aguja == null)
			return -1;

		for (int i = 0; i < pajar.getLength(); ++i) {
			if (aguja.equals(pajar.item(i))) {
				return i;
			}
		}
		return -1;
	}

	/**
	 * Cierra el modelo de presentación del documento, interrumpiendo cualquier
	 * proceso que pueda quedarse ejecutándose indefinidamente de lo contrario.
	 */
	public void cerrar() {
		notCambios.interrumpir();
	}
}
