package es.uca.xmleye.util;

/*
 Copyright (C) 2005-2008 Antonio García Domínguez

 This program is free software; you can redistribute it 
 and/or modify it under the terms of the GNU General Public 
 License as published by the Free Software Foundation; 
 either version 2 of the License, or (at your option) any 
 later version.

 This program is distributed in the hope that it will be 
 useful, but WITHOUT ANY WARRANTY; without even the implied 
 warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 PURPOSE. See the GNU General Public License for more 
 details.

 You should have received a copy of the GNU General Public 
 License along with this program; if not, write to the Free 
 Software Foundation, Inc., 59 Temple Place, Suite 330, 
 Boston, MA 02111-1307 USA
 */

import java.io.*;
import java.util.*;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executors;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;

/**
 * <p>
 * Envoltorio para implementar el uso de un proceso como un filtro, redirigiendo
 * su stderr al de la JVM.
 * </p>
 * 
 * <p>
 * Esta clase está dirigida especialmente a filtros, pero también se puede usar
 * como una clase de conveniencia, que vuelca la salida de errores a la de la
 * JVM y devuelve el valor de retorno del proceso de forma sencilla.
 * </p>
 */
public class ProcesoFiltro {
	private TareaSalida tareaEntrada;
	private TareaEntrada tareaSalida;
	private TareaEntrada tareaError;

	private ProcessBuilder procBuilder;
	private static ExecutorService executor = Executors.newFixedThreadPool(3);

	/**
	 * Crea un proceso con filtrado.
	 * 
	 * Se usará la orden <code>orden</code> y se ejecutarán las tareas
	 * especificadas que no sean <code>null</code> sobre los flujos del
	 * proceso.
	 * 
	 * @param orden
	 *            Orden a ejecutar.
	 * @param tareaEntrada
	 *            Tarea a ejecutar sobre stdin del proceso, posiblemente
	 *            <code>null</code>.
	 * @param tareaSalida
	 *            Tarea a ejecutar sobre stdout del proceso, posiblemente
	 *            <code>null</code>.
	 * @param tareaError
	 *            Tarea a ejecutar sobre stderr del proceso, posiblemente
	 *            <code>null</code>.
	 */
	public ProcesoFiltro(ArrayList<String> orden, TareaSalida tareaEntrada,
			TareaEntrada tareaSalida, TareaEntrada tareaError) {
		this.tareaEntrada = tareaEntrada;
		this.tareaSalida = tareaSalida;
		this.tareaError = tareaError;

		// Creamos el proceso
		this.procBuilder = new ProcessBuilder(orden);
	}

	/**
	 * <p>
	 * Lanza el filtro. Además de volcar la entrada en la stdin del proceso y su
	 * stdout a la salida, redirige stderr del proceso a la stderr de la JVM.
	 * </p>
	 * 
	 * <p>
	 * La ejecución de este método no termina hasta que se haya leído toda la
	 * salida.
	 * </p>
	 * 
	 * @param matarSiInterrumpido
	 *            Si <code>true</code>, el proceso se matará si se interrumpe
	 *            la espera a su resultado.
	 * @throws IOException
	 *             Error de E/S al lanzar el proceso.
	 * @throws InterruptedException
	 *             Se interrumpió la ejecución del método antes de que se
	 *             pudiera obtener su valor de retorno.
	 * @return Valor de retorno de la orden ejecutada (0 si todo va bien,
	 *         distinto de 0 si algo va mal).
	 * @throws ExecutionException 
	 */
	@SuppressWarnings("unchecked")
	public int iniciar(boolean matarSiInterrumpido) throws IOException,
			InterruptedException, ExecutionException {
		final Process process = procBuilder.start();
		Future resEntrada = null;
		Future resSalida = null;
		Future resError = null;

		if (tareaEntrada != null) {
			OutputStream stdin = new BufferedOutputStream(process
					.getOutputStream());
			tareaEntrada.setFlujoSalida(stdin);
			resEntrada = executor.submit((Callable<Integer>) tareaEntrada);
		}

		if (tareaSalida != null) {
			InputStream stdout = new BufferedInputStream(process
					.getInputStream());
			tareaSalida.setFlujoEntrada(stdout);
			resSalida = executor.submit((Callable<Integer>) tareaSalida);
		}

		if (tareaError != null) {
			InputStream stderr = new BufferedInputStream(process
					.getErrorStream());
			tareaError.setFlujoEntrada(stderr);
			resError = executor.submit((Callable<Integer>) tareaError);
		}

		try {
			int codigo = process.waitFor();
			
			// Esperamos a que todos los flujos se cierren
			if (resEntrada != null)
				resEntrada.get();
			if (resSalida != null)
				resSalida.get();
			if (resError != null)
				resError.get();
			
			return codigo;
		} finally {
			if (resEntrada != null && !resEntrada.isDone())
				resEntrada.cancel(true);
			if (resSalida != null && !resSalida.isDone())
				resSalida.cancel(true);
			if (resError != null && !resError.isDone())
				resError.cancel(true);

			if (matarSiInterrumpido) {
				process.destroy();
			}
		}
	}

}
